#include <stdio.h>
#include <string.h>
#include <time.h>
#include <sys/time.h>

typedef unsigned char u8;
typedef unsigned short int u16;
typedef unsigned int u32;
typedef unsigned long long u64;

#define rdtsc(low,high) \
     __asm__ __volatile__("rdtsc" : "=a" (low), "=d" (high))

static inline unsigned long long rdtsc_read(void)
{
	unsigned long high, low;
	rdtsc(low,high);
	return ((unsigned long long)high << 32) | low;
}

static inline unsigned long long
timespec_diff(struct timespec *start, struct timespec *end)
{
	unsigned long long res;
	
	if (end->tv_nsec < start->tv_nsec) {
		res = (end->tv_sec - start->tv_sec - 1) * 1000000000LL;
		res += 1000000000 + end->tv_nsec - start->tv_nsec;
	} else {
		res = (end->tv_sec - start->tv_sec) * 1000000000LL;
		res += end->tv_nsec - start->tv_nsec;
	}
	return res;
}

static inline u32 rol32(u32 word, unsigned int shift)
{
	return (word << shift) | (word >> (32 - shift));
}

struct __una_u32 { u32 x __attribute__((packed)); };

static inline u32 __get_unaligned_cpu32(const void *p)
{
	const struct __una_u32 *ptr = (const struct __una_u32 *)p;
	return ptr->x;
}

#include "jhash2.h"

static u32 jhash2_fn(const void *key, u32 length, u32 initval)
{
	return jhash(key, length, initval);
}

static u32 jhash2w_fn(const u32 *k, u32 length, u32 initval)
{
	return jhash2(k, length, initval);
}

#include "jhash3.h"

static u32 jhash3_fn(const void *key, u32 length, u32 initval)
{
	return jhash3(key, length, initval);
}

static u32 jhash3w_fn(const u32 *k, u32 length, u32 initval)
{
	return jhash32(k, length, initval);
}

enum {
	HASH_JHASH2 = 0,
	HASH_JHASH3,
	HASH_MAX
};

struct hashfn {
	const char name[16];
	u32(*hashfn) (const void * k, u32 length, u32 initval);
	u32(*hashfn2) (const u32 * k, u32 length, u32 initval);
};

struct hashfn hashes[HASH_MAX] = {
	[HASH_JHASH2] = {
		.name = "jhash2",
		.hashfn = jhash2_fn,
		.hashfn2 = jhash2w_fn,
	},
	[HASH_JHASH3] = {
		.name = "jhash3",
		.hashfn = jhash3_fn,
		.hashfn2 = jhash3w_fn,
	},
};

#define REPEAT	1000000
#define MAXKEY	64

u32 sizes[] = {1, 2, 3, 4, 8, 16, 32, MAXKEY};

#define SIZES	(sizeof(sizes)/sizeof(1))

u32 key[MAXKEY];

#ifdef RDTSC
#define START_TIMER	t = rdtsc_read()
#define END_TIMER	e += rdtsc_read() - t
#else
#define START_TIMER	\
        clock_gettime(CLOCK_PROCESS_CPUTIME_ID, &start)
#define	END_TIMER	\
	clock_gettime(CLOCK_PROCESS_CPUTIME_ID, &end);	\
        e += timespec_diff(&start, &end)
#endif

void driver(const struct hashfn *hash, long double *elapsed)
{
	u32 i, j, k;
	u32 h = 0x12345678;
	u64 e, t;
	struct timespec start, end;
	
	for (k = 0; k < SIZES; k++) {
		e = 0;
		START_TIMER;
		for (i = 0; i < REPEAT; i++) {
			h = hash->hashfn((void *)key, sizes[k]*sizeof(u32), h);
		}
		END_TIMER;
		elapsed[k] = ((long double)e)/REPEAT;
		printf("%s (%d bytes), elapsed %llu (%llf)\n",
		       hash->name, sizes[k]*sizeof(u32), e, elapsed[k]);
	}
}

void driver2(const struct hashfn *hash, long double *elapsed)
{
	u32 i, j, k;
	u32 h = 0x12345678;
	u64 t, e;
	struct timespec start, end;
	
	for (k = 0; k < SIZES; k++) {
		e = 0;
		START_TIMER;
		for (i = 0; i < REPEAT; i++) {
			h = hash->hashfn2((u32 *)key, sizes[k], h);
		}
		END_TIMER;
		elapsed[k] = ((long double)e)/REPEAT;
		printf("%s (%d words), elapsed %llu (%llf)\n",
		       hash->name, sizes[k], e, elapsed[k]);
	}
}

int main()
{
	int i, j, k;
	long double elapsed[HASH_MAX][SIZES];

	memset(key, 'x', sizeof(key));
	
	for (i = 0; i < HASH_MAX; i++)
		driver(&hashes[i], elapsed[i]);

	for (i = 0; i < HASH_MAX; i++)
		for (j = i + 1; j < HASH_MAX; j++)
			for (k = 0; k < SIZES; k++)
				printf("%s/%s (%u bytes): %Lf\n",
				       hashes[i].name, hashes[j].name,
				       sizes[k]*sizeof(u32),
				       elapsed[i][k]/elapsed[j][k]);

	for (i = 0; i < HASH_MAX; i++)
		driver2(&hashes[i], elapsed[i]);
	for (i = 0; i < HASH_MAX; i++)
		for (j = i + 1; j < HASH_MAX; j++)
			for (k = 0; k < SIZES; k++)
				printf("%s/%s (%u words): %Lf\n",
				       hashes[i].name, hashes[j].name,
				       sizes[k],
				       elapsed[i][k]/elapsed[j][k]);

	return 0;
}
